import abc
import io
import json as libjson
from urllib.parse import urlencode

from .response import Response
from .helpers import encode_multipart_data


class WSGIResponse(Response):
    def write(self, data):
        if self.body is None:
            self.body = b''

        if isinstance(data, str):
            self.body += data.encode()
        else:
            self.body += data


class Connector(metaclass=abc.ABCMeta):
    def request(self, verb='GET', url='/', form=None, multipart=None,
                json=None, environ=None, headers=None, body=None, **kw):
        headers = headers or []

        if body is None:
            if multipart:
                content_type, body, content_length = \
                    encode_multipart_data(multipart)
                headers.append(('Content-Type', content_type))
                headers.append(('Content-Length', str(content_length)))

            elif json:
                body = libjson.dumps(json)
                headers.append(
                    ('Content-Type', 'application/json;charset:utf-8')
                )
                headers.append(('Content-Length', str(len(body))))

            elif isinstance(form, dict):
                body = urlencode(form)
                headers.append(
                    ('Content-Type', 'application/x-www-form-urlencoded')
                )

        if isinstance(body, str):
            body = body.encode()

        return self._send_request(verb, url, environ, headers, body)

    @abc.abstractmethod
    def _send_request(self, verb, url, environ, headers, body=None, **kw):
        pass


class WSGIConnector(Connector):
    def __init__(self, application, environ=None):
        self.application = application
        self.environ = environ

    def _prepare_environ(self, verb, url, headers, form=None,
                         extra_environ=None):
        if isinstance(form, io.BytesIO):
            input_file = form
        elif form:
            input_file = io.BytesIO(form)
            input_file.seek(0)
        else:
            input_file = io.BytesIO()

        error_file = io.StringIO()
        environ = self.environ.copy() if self.environ else {}
        environ['wsgi.input'] = input_file
        environ['wsgi.errors'] = error_file
        environ['wsgi.version'] = (1, 0)
        environ['wsgi.multithread'] = False
        environ['wsgi.multiprocess'] = False
        environ['wsgi.run_once'] = True
        environ['wsgi.url_scheme'] = 'http'
        environ['REQUEST_METHOD'] = verb

        if '?' in url:
            url, query = url.split('?', 1)
        else:
            query = ''
        environ['QUERY_STRING'] = query
        environ['PATH_INFO'] = url

        if extra_environ:
            environ.update(extra_environ)

        for k, v in headers:
            key = k.upper().replace('-', '_')
            if key not in ['CONTENT_TYPE', 'CONTENT_LENGTH']:
                key = 'HTTP_' + key
            environ[key] = v

        return environ

    def _send_request(self, verb, url, environ, headers, body=None, **kw):
        environ_ = self._prepare_environ(verb, url, headers, body, environ)
        response = None

        def start_response(status, headers, exc_info=None):
            nonlocal response
            if exc_info:
                raise (exc_info[0], exc_info[1], exc_info[2])

            response = WSGIResponse(status, headers)
            return response.write

        result = self.application(
            environ_,
            start_response
        )

        try:
            for i in result:
                response.write(i)
        finally:
            if hasattr(result, 'close'):
                result.close()

        return response




import os
import io
import base64
from urllib.parse import parse_qs
from mimetypes import guess_type


def normalize_query_string(query):
    if not query:
        return None
    return {
        k: v[0] if len(v) == 1 else v for k, v in parse_qs(query).items()
    } if isinstance(query, str) else query


def encode_multipart_data(fields):
    boundary = ''.join(
        ['-----', base64.urlsafe_b64encode(os.urandom(27)).decode()]
    )
    crlf = b'\r\n'
    lines = []

    for key, value in fields.items():
        if not hasattr(value, 'read'):
            lines.append('--' + boundary)
            lines.append('Content-Disposition: form-data; name="%s"' % key)
            lines.append('')
            if not isinstance(value, str):
                value = str(value)
            lines.append(value)
        else:
            filename = value.name if hasattr(value, 'name') else key
            lines.append('--' + boundary)
            lines.append(
                'Content-Disposition: form-data; name="%s"; filename="%s"' %
                (key, filename))
            lines.append(
                'Content-Type: %s' %
                (guess_type(filename)[0] or 'application/octet-stream'))
            lines.append('')
            lines.append(value.read())

    lines.append('--' + boundary + '--')
    lines.append('')

    body = io.BytesIO()
    length = 0
    for l in lines:
        line = (l if isinstance(l, bytes) else l.encode()) + crlf
        length += len(line)
        body.write(line)
    body.seek(0)
    content_type = 'multipart/form-data; boundary=%s' % boundary
    return content_type, body, length






import cgi
import json

from bddrest import Given, response, status, when


def wsgi_application(environ, start_response):
    if environ.get('CONTENT_TYPE', '').startswith('application/json'):
        result = json.loads(environ['wsgi.input'].read())
    else:
        form = cgi.FieldStorage(
            fp=environ['wsgi.input'],
            environ=environ,
            strict_parsing=False,
            keep_blank_values=True
        )
        result = {}
        for k in form.keys():
            v = form[k]
            value = v.value

            result[k] = value.decode() if isinstance(value, bytes) else value

    start_response('200 OK', [
        ('Content-Type', 'application/json;charset=utf-8'),
    ])
    yield json.dumps(result).encode()


def test_none_parameters_within_form():
    form = dict(a=None)

    with Given(
        wsgi_application,
        title='test none parameters in a url-encoded form',
        verb='POST',
        form=form
    ):
        assert status == 200
        assert response.json['a'] == 'None'

    with Given(
        wsgi_application,
        title='test none parameters in a json form',
        verb='POST',
        json=form
    ):
        assert status == 200
        assert response.json['a'] is None

    with Given(
        wsgi_application,
        title='test none parameters in a multipart form',
        verb='POST',
        multipart=form
    ):
        assert status == 200
        assert response.json['a'] == 'None'






from .headerset import HeaderSet
from .response import HTTPStatus, Response
from .specification import Call, FirstCall, AlteredCall
from .authoring import Given, when, story, response, Story, Append, Update, \
    Remove, status, given_form
from .exceptions import InvalidUrlParametersError, CallVerifyError


__version__ = '1.15.3b'




import io
import os
import sys
import json
import cgi

from bddrest.response import Response
from bddrest.connectors import WSGIConnector


def test_wagi_streaming():
    def app(environ, start_response):
        write = start_response('200 Ok', [('Content-Type', 'aplication/json')])
        write(b'abc\n')
        return []

    connector = WSGIConnector(app)
    response = connector.request()
    assert isinstance(response, Response)
    assert response.text == 'abc\n'


def test_wsgi_yield():
    def app(environ, start_response):
        start_response('200 Ok', [('Content-Type', 'aplication/json')])
        yield 'abc\n'

    connector = WSGIConnector(app)
    response = connector.request()
    assert response.text == 'abc\n'


def test_wsgi_forms():
    def app(environ, start_response):
        if environ.get('CONTENT_TYPE', '').startswith('application/json'):
            result = json.loads(environ['wsgi.input'].read())
        else:
            form = cgi.FieldStorage(
                fp=environ['wsgi.input'],
                environ=environ,
                strict_parsing=False,
                keep_blank_values=True
            )
            result = {}
            for k in form.keys():
                v = form[k]
                value = v.value

                result[k] = \
                    value.decode() if isinstance(value, bytes) else value

        start_response('200 OK', [
            ('Content-Type', 'application/json;charset=utf-8'),
        ])
        yield json.dumps(result).encode()

    connector = WSGIConnector(app)

    # Simple string as request body with JSON content type
    response = connector.request(
        'POST',
        headers=[('Content-Type', 'application/json')],
        body='{"a": "b"}'
    )
    assert response.json.items() == dict(a='b').items()

    # JSON form
    response = connector.request(
        'POST',
        json=dict(a='b')
    )
    assert response.json.items() == dict(a='b').items()

    # url encoded:
    response = connector.request('POST', form=dict(a=1, b=2))
    assert response.json.items() == dict(a='1', b='2').items()

    # multipart
    form = {
        'a': 1,
        'b': io.StringIO('Hello')
    }
    response = connector.request('POST', multipart=form)
    assert response.json.items() == dict(a='1', b='Hello').items()


def test_query_string():
    def app(environ, start_response):
        query = environ['QUERY_STRING']
        start_response('200 OK', [
            ('Content-Type', 'text/plain'),
        ])
        yield query

    connector = WSGIConnector(app)

    response = connector.request(
        'POST',
        '/?a=b'
    )

    assert response.text == 'a=b'

